@@ -988,7 +988,8 @@ static int server_connect(void) {
   while (connect(sfd,(struct sockaddr*)&ssockname,sizeof(ssockname))) {
     if (errno == ECONNREFUSED || errno == ENOENT)
       syscallerror("uservd daemon is not running - service not available");
-    syscallerror("unable to connect to uservd daemon");
+    if (errno != EINTR)
+      syscallerror("unable to connect to uservd daemon: %m");
   }
 
   return sfd;
